rm(list=ls())
library(marima2)
sapply(dir("functions/", full.names=TRUE), source)


## ---------------------------------------------------------------------------------------
# A simulation function for ARMA simulation, use model as arima.sim, i.e. flip sign of phi (into ar) coefficients
simarmax <- function(model, x, sd=1, nburnin=100){
  n <- length(x) + nburnin
  x <- c(rep(x[1],nburnin), x)
  # Take the ar and ma part
  ar <- model$ar
  ma <- model$ma
  xcoef <- model$x
  # The order (i.e. the number of lags)
  p <- length(ar)
  q <- length(ma)
  r <- length(xcoef)
  # The vector for the simulation result
  y <- numeric(n)
  # Generate the random normal values
  eps <- rnorm(n, sd=sd)
  # Run the simulation
  for(i in (max(p,q,r)+1):n){
    y[i] <- eps[i] + sum(ar * y[i-(1:p)]) + sum(xcoef * x[i-(1:r)] ) + sum(ma * eps[i-(1:q)])
  }
  # Return without the burn-in period
  return(y[(nburnin+1):n])
}


## ---------------------------------------------------------------------------------------
# Simulate an ARMAX(1,1) model, with a step input
# Test it by comparing
n <- 100
# Make a step input
x <- c(rep(0,50), rep(10,n-50))
model <- list(ar=c(0.7), ma=c(1), x=c(0.9))
set.seed(12)
y <- simarmax(model, x)
# Plot to see it
plot(y, type="b")
lines(x, type="s", col=2)
legend("topleft", c("Input","Output"), col=2:1, lty=1)

## ---------------------------------------------------------------------------------------
# Estimate with arima (must lag x1)
x1 <- lagvec(x,1)
x1[1] <- 0
(fitArima <- arima(y, order=c(1,0,1), xreg=x1, include.mean=FALSE))
# Validation ok?
residArima <- validate(fitArima)

# Estimate with marima
fitMarima <- marima("y ~ AR(1) + x(1) + MA(1)", data=data.frame(y=y,x=x))
summary(fitMarima)
# Validation ok?
residMarima <- validate(fitMarima)
# See the prediction in the training period

# See the prediction in the training period
plot(y-residArima, type="l")
lines(y-residMarima)

# See the parameter estimates
fitArima
summary(fitMarima)

# Difference in residuals
plot(residArima - residMarima)


## ---------------------------------------------------------------------------------------
# CHALLENGE: CAN YOU FIND A WAY TO FIT AN ARMAX!? (like in marima)
## ---------------------------------------------------------------------------------------

## ---------------------------------------------------------------------------------------
# First an AR model with lm
# Add the lags to X
maxlag <- 5
D <- data.frame(y=y, x=x)
for(i in 1:maxlag){
  tmp <- lagdf(D[ ,1:2], i)
  names(tmp) <- paste0(names(tmp),".l",i)
  D <- cbind(D, tmp)
}
D <- D[-1:-maxlag, ]
row.names(D) <- 1:nrow(D)
names(D)


## ---------------------------------------------------------------------------------------
# TASK: Try making the ARMAX fit marima style with lm()
## ---------------------------------------------------------------------------------------

# First fit an ARX with lm()
fit <- lm(y ~ 0 + y.l1 + x.l1, data=D)
residlm <- fit$residuals
plot(residlm, type="l", col=2)

# Then repeat using the residuals from last iteration
Resid <- list()
resid <- residlm
for(i in 1:100){
  D$resid.l1 <- lagvec(resid, 1)
  fit2 <- lm(y ~ 0 + y.l1 + x.l1 + resid.l1, data=D)
  resid[as.integer(names(fit2$residuals))] <- fit2$residuals
  #validate(fit2)
  lines(resid, type="l", col=i)
  Resid[[i]] <- resid
}

for(i in 2:100){
  print(sum(Resid[[i-1]] - Resid[[i]])^2)
}


## ---------------------------------------------------------------------------------------
# AIC example (MOVE THIS TO ANOTHER FILE)
## ---------------------------------------------------------------------------------------
#download.file("https://02417.compute.dtu.dk/material/marima2_0.1.tar.gz", "marima2_0.1.tar.gz")
#install.packages("marima2_0.1.tar.gz", repos=NULL)
library(marima2)

## ---------------------------------------------------------------------------------------
# A simulation function for ARMA simulation, use model as arima.sim, i.e. flip sign of phi (into ar) coefficients
simarmax <- function(model, x, sd=1, nburnin=100){
  n <- length(x) + nburnin
  x <- c(rep(x[1],nburnin), x)
  # Take the ar and ma part
  ar <- model$ar
  ma <- model$ma
  xcoef <- model$x
  # The order (i.e. the number of lags)
  p <- length(ar)
  q <- length(ma)
  r <- length(xcoef)
  # The vector for the simulation result
  y <- numeric(n)
  # Generate the random normal values
  eps <- rnorm(n, sd=sd)
  # Run the simulation
  for(i in (max(p,q,r)+1):n){
    y[i] <- eps[i] + sum(ar * y[i-(1:p)]) + sum(xcoef * x[i-(1:r)] ) + sum(ma * eps[i-(1:q)])
  }
  # Return without the burn-in period
  return(y[(nburnin+1):n])
}
lagdf <- function(x, lag) {
  # Lag x, i.e. delay x lag steps
  if (lag > 0) {
    x[(lag+1):nrow(x), ] <- x[1:(nrow(x)-lag), ]
    x[1:lag, ] <- NA
  }else if(lag < 0) {
    # Lag x "ahead in time"
    x[1:(nrow(x)-abs(lag)), ] <- x[(abs(lag)+1):nrow(x), ]
    x[(nrow(x)-abs(lag)+1):nrow(x), ] <- NA
  }
  return(x)
}


## ---------------------------------------------------------------------------------------
# Simulate an ARMAX(1,1) model, with a step input
# Test it by comparing
n <- 100
# Make a step input
x <- c(rep(0,50), rep(10,n-50))
model <- list(ar=c(0.6), ma=c(1), x=c(0.9))
set.seed(12)
y <- simarmax(model, x)
# Plot to see it
plot(y, type="b")
lines(x, type="s")


# Estimate with marima
fitMarima <- marima("y ~ AR(1) + x(1) + MA(1)", data=data.frame(y=y,x=x))

# Estimate with lm
# Add the lags to X
maxlag <- 5
D <- data.frame(y=y, x=x)
for(i in 1:maxlag){
  tmp <- lagdf(D[ ,1:2], i)
  names(tmp) <- paste0(names(tmp),".l",i)
  D <- cbind(D, tmp)
}
names(D)
fit <- lm(y ~ 0 + y.l1 + x.l1, data=D)


## ---------------------------------------------------------------------------------------
# AIC example
## ---------------------------------------------------------------------------------------
# The log likelihood for lm() fit
resid <- fit$residuals
# To be very correct the standard deviation needs to be calculated in a slightly different
sd(resid)
sqrt(sum((resid - mean(resid))^2) / (length(resid)-1) )
(sd <- sqrt(sum((resid - mean(resid))^2) / length(resid) ))
# The log-likelihood
loglik <- sum(log(dnorm(resid, sd=sd)))
# AIC = penalty*p - 2*log(likelihood)
# p is the number of parameters, must add one for least squares, because the standard deviation of the error is not directly estimated (it's used in the loglikelihood calculation above)
# Check
logLik(fit)
loglik
# and get the number of parameters
p <- length(fit$coefficients) + 1
AIC(fit)
2*p - 2*loglik
# and
BIC(fit)
log(length(resid))*p - 2*loglik


# For MARIMA
resid <- fitMarima$residuals[1, ]
resid <- resid[!is.na(resid)]
sd <- sqrt(sum((resid - mean(resid))^2) / length(resid) )
loglik <- sum(log(dnorm(resid, sd=sd)))
loglik
# Find the number of parameters
p <- sum(fitMarima$ar.estimates[ , ,-1] != 0) + sum(fitMarima$ma.estimates[ , ,-1] != 0)
p <- p + 1
# AIC
2*p - 2*loglik
# BIC
log(length(resid))*p - 2*loglik


## ---------------------------------------------------------------------------------------
# In order to use AIC, be careful, because higher order models have more initial NAs in the residuals! so should have a "burn-in" to compare
# ARX first order
fit1 <- lm(y ~ 0 + y.l1 + x.l1, data=D)
length(fit1$residuals)
# ARX second order
fit2 <- lm(y ~ 0 + y.l1 + y.l2 + x.l1 + x.l2, data=D)
length(fit2$residuals)

# So we should have a burn-in, e.g. of 5 steps (in priciple the max order of the models)
score <- function(resid, n, p, nburnin, penalty=2){
  # How many to use
  nuse <- n - nburnin
  # Take the nuse last values
  resid <- resid[(length(resid)-nuse+1):length(resid)]
  # Calc
  sd <- sqrt(sum((resid - mean(resid))^2) / length(resid) )
  loglik <- sum(log(dnorm(resid, sd=sd)))
  # Return
  penalty*p - 2*loglik
}  
# AIC
score(fit$residuals, n=nrow(D), p=length(fit$coefficients)+1, nburnin=0)
AIC(fit)

# To compare models of different order use a burn-in period
# With a burn-in period of 5 (simply just excluding first nburnin residuals)
score(fit1$residuals, n=nrow(D), p=length(fit1$coefficients)+1, nburnin=5)
score(fit2$residuals, n=nrow(D), p=length(fit2$coefficients)+1, nburnin=5)

# Likewise for marima:
fitMarima1 <- marima("y ~ AR(1) + x(1) + MA(1)", data=data.frame(y=y,x=x))
fitMarima2 <- marima("y ~ AR(1:2) + x(1:2) + MA(1:2)", data=data.frame(y=y,x=x))
#
p <- sum(fitMarima1$ar.estimates[ , ,-1] != 0) + sum(fitMarima1$ma.estimates[ , ,-1] != 0)
(p <- p + 1)
score(fitMarima1$residuals, n=nrow(D), p=p, nburnin=5)
#
p <- sum(fitMarima2$ar.estimates[ , ,-1] != 0) + sum(fitMarima2$ma.estimates[ , ,-1] != 0)
(p <- p + 1)
score(fitMarima2$residuals, n=nrow(D), p=p, nburnin=5)
